"""The names :py:class:`.render.Color`, :py:class:`.render.Material`,
:py:class:`.render.Renderer`, :py:class:`.wrapper.NTracer`,
:py:data:`.wrapper.CUBE`, :py:data:`.wrapper.SPHERE` and
:py:func:`.kdtree_builder.build_composite_scene` are imported here for
convenience."""


from ntracer.render import Color,Material,Renderer
from ntracer.wrapper import NTracer,CUBE,SPHERE
from ntracer.kdtree_builder import build_composite_scene

